//---------------------------------------------------------------------------------
// File Name: RaceTrack.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "RaceTrack.h"

#include "../TA/Common/Geometry.h"
#include "../TA/Common/Error.h"

#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/Physics.h"

#include "../Interface/IMesh.h"
#include "../Interface/IApp.h"

#include <memory.h>

#include <stdio.h>


using namespace TA;

const int k_nNumCornerSegs = 10;
const float k_fCornerCurve = 1.25f;

static SurfaceAttribute s_surfaceAttribute;
static int s_nSurfaceAttribute = 0;


//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
RaceTrack::RaceTrack()
{
	m_pMesh = 0;
	m_pStaticObject = 0;
	m_pStraightArray = 0;
	m_nStraigtArraySize = 0;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
RaceTrack::~RaceTrack()
{
	Finalise();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RaceTrack::Initialise(IApp* pApp, const char* szFileName)
{
	m_pMesh = 0;
	CollisionObjectAABBMesh* pCollisionObjectAABBMesh = 0;

	s_surfaceAttribute.SetCollisionFriction(1.0f);
	s_nSurfaceAttribute = Physics::GetInstance().AddSurfaceAttribute(&s_surfaceAttribute);

	const Vec3 k_v3LightDir = Vec3(1.0f, 1.0f, 1.0f).GetNormal();

	const float k_fScale = 1.5f;
	const float k_fTrackWidth = 50.0f;

	FILE* pFile = fopen(szFileName, "r");	
	if (!pFile)
	{
		char szError[2048];
		_snprintf(szError, 2048, "Error loading race track: Couldn't find %s", szFileName);
		TA_ERROR_REPORT_SEVERE(szError);
		Finalise();
		return;
	}
	int nShowWalls = 0;
	int nScanCount = fscanf(pFile, " %d", &nShowWalls);	
	if (nScanCount != 1)
	{		
		fclose(pFile);
		char szError[2048];
		_snprintf(szError, 2048, "Error loading race track: Invalid data in file %s", szFileName);
		TA_ERROR_REPORT_SEVERE(szError);
		Finalise();
		return;
	}
	int nNumSegs = 0;
	nScanCount = fscanf(pFile, " %d", &nNumSegs);	
	if (nScanCount != 1)
	{		
		fclose(pFile);
		char szError[2048];
		_snprintf(szError, 2048, "Error loading race track: Invalid data in file %s", szFileName);
		TA_ERROR_REPORT_SEVERE(szError);
		Finalise();
		return;
	}
	Vec3* pv3PointArray = new Vec3[nNumSegs];
	memset(pv3PointArray, 0, sizeof (Vec3) * nNumSegs);
	for (int nPoint = 0; nPoint < nNumSegs; nPoint++)
	{
		int nScanCount = fscanf(
			pFile, 
			" %f %f %f", 
			&pv3PointArray[nPoint].x, 
			&pv3PointArray[nPoint].y, 
			&pv3PointArray[nPoint].z);	
		if (nScanCount != 3)
		{		
			fclose(pFile);
			char szError[2048];
			_snprintf(szError, 2048, "Error loading race track: Invalid data in file %s", szFileName);
			TA_ERROR_REPORT_SEVERE(szError);
			Finalise();
			return;
		}
	}

	fclose(pFile);
	pFile = 0;

	const float k_fCornerScale = 40.0f;
	TA_ASSERT(m_pStraightArray == 0);
	m_pStraightArray = new Straight[nNumSegs];
	m_nStraigtArraySize = nNumSegs;

	for (int nStraight = 0; nStraight < nNumSegs; nStraight++)
	{
		const Vec3& v3PosA = pv3PointArray[nStraight];
		const Vec3& v3PosB = nStraight + 1 < nNumSegs ?
			pv3PointArray[nStraight + 1] : pv3PointArray[0];
		const Vec3& v3PosC = nStraight + 2 < nNumSegs ?
			pv3PointArray[nStraight + 2] : pv3PointArray[nStraight + 2 - nNumSegs];
		const Vec3& v3PosD = nStraight - 1 >= 0 ?
			pv3PointArray[nStraight - 1] : pv3PointArray[nStraight - 1 + nNumSegs];

		Vec3 v3Normal = v3PosB - v3PosA;
		TA_ASSERT(v3Normal.GetMagnitude() > 0.0f);
		v3Normal.Normalise();

		Vec3 v3NextNormal = v3PosC - v3PosB;
		TA_ASSERT(v3NextNormal.GetMagnitude() > 0.0f);
		v3NextNormal.Normalise();


		Straight& straight = m_pStraightArray[nStraight];
		float fCornerScaleMod = (1.1f - v3NextNormal.Dot(v3Normal)) * k_fCornerScale;
		TA_ASSERT(fCornerScaleMod > 0.0f);
		if (fCornerScaleMod > k_fCornerScale)
			fCornerScaleMod = k_fCornerScale;


		straight.v3End = v3PosB * k_fScale - v3Normal * fCornerScaleMod;

		Vec3 v3PrevNormal = v3PosA - v3PosD;
		TA_ASSERT(v3PrevNormal.GetMagnitude() > 0.0f);
		v3PrevNormal.Normalise();

		fCornerScaleMod = (1.1f - v3PrevNormal.Dot(v3Normal)) * k_fCornerScale;
		TA_ASSERT(fCornerScaleMod > 0.0f);
		if (fCornerScaleMod > k_fCornerScale)
			fCornerScaleMod = k_fCornerScale;

		straight.v3Start = v3PosA * k_fScale + v3Normal * fCornerScaleMod;
	}

	if (nShowWalls)
	{
	
		int nNumCollisionVertices = 
			nNumSegs * 4 * k_nNumCornerSegs * 4;
		int nNumRenderVertices = 
			nNumSegs * 4 * k_nNumCornerSegs * 6;
		int nNumCollisionPolygons = 
			nNumSegs * 2 * k_nNumCornerSegs * 6;
		TA_ASSERT(pCollisionObjectAABBMesh == 0);
		pCollisionObjectAABBMesh = CollisionObjectAABBMesh::CreateNew();
		pCollisionObjectAABBMesh->Initialise(
			nNumCollisionVertices,
			nNumCollisionPolygons,
			nNumCollisionPolygons * 4);
		
		TA_ASSERT(m_pMesh == 0);
		m_pMesh = pApp->CreateMesh();
		m_pMesh->Initialise(
			nNumRenderVertices,
			nNumCollisionPolygons * 12,
			TA_CHAR("Assets/Textures/Track.tga"));
		
		for (nStraight = 0; nStraight < nNumSegs; nStraight++)
		{
			Straight& straight = m_pStraightArray[nStraight];
			
			Straight& straightA = m_pStraightArray[nStraight];
			Straight& straightB = nStraight + 1 < nNumSegs ? 
				m_pStraightArray[nStraight + 1] : m_pStraightArray[0];

			Vec3 v3VelA = straightA.v3End - straightA.v3Start;
			v3VelA.Normalise();
			Vec3 v3NormalA = v3VelA.CrossWithUnitY();
			v3NormalA.Normalise();

			Vec3 v3VelB = straightB.v3End - straightB.v3Start;
			v3VelB.Normalise();
			Vec3 v3NormalB = v3VelB.CrossWithUnitY();
			v3NormalB.Normalise();

			Vec3 v3Start[4];
			v3Start[0] = straightA.v3End + v3NormalA * (0.5f * k_fTrackWidth);
			v3Start[1] = straightA.v3End + v3NormalA * (0.54f * k_fTrackWidth);
			v3Start[2] = straightA.v3End - v3NormalA * (0.54f * k_fTrackWidth);
			v3Start[3] = straightA.v3End - v3NormalA * (0.5f * k_fTrackWidth);

			Vec3 v3End[4];
			v3End[0] = straightB.v3Start + v3NormalB * (0.5f * k_fTrackWidth);
			v3End[1] = straightB.v3Start + v3NormalB * (0.54f * k_fTrackWidth);
			v3End[2] = straightB.v3Start - v3NormalB * (0.54f * k_fTrackWidth);
			v3End[3] = straightB.v3Start - v3NormalB * (0.5f * k_fTrackWidth);

			Vec3 v3End2[4];
			v3End2[0] = straightB.v3End + v3NormalB * (0.5f * k_fTrackWidth);
			v3End2[1] = straightB.v3End + v3NormalB * (0.54f * k_fTrackWidth);
			v3End2[2] = straightB.v3End - v3NormalB * (0.54f * k_fTrackWidth);
			v3End2[3] = straightB.v3End - v3NormalB * (0.5f * k_fTrackWidth);

			float fCurve[4];
			fCurve[0] = (v3End[0] - v3Start[0]).GetMagnitude() * k_fCornerCurve;
			fCurve[1] = (v3End[1] - v3Start[1]).GetMagnitude() * k_fCornerCurve;
			fCurve[2] = (v3End[2] - v3Start[2]).GetMagnitude() * k_fCornerCurve;
			fCurve[3] = (v3End[3] - v3Start[3]).GetMagnitude() * k_fCornerCurve;
		

			for (int nSeg = 0; nSeg < k_nNumCornerSegs * 2; nSeg++)
			{
				for (int i = 0; i < 4; i++)
				{
					const Vec3 v3BottomOffset = - k_v3UnitY * 5.0f;

					Vec3 v3Pos;
					Vec3 v3LightNormal;
					if (nSeg < k_nNumCornerSegs)
					{			
						v3Pos = Geometry::CubicInterpolate(
							v3Start[i], 
							v3End[i], 
							v3VelA * fCurve[i], 
							v3VelB * fCurve[i], 
							(float)nSeg / (float)k_nNumCornerSegs);	

						Vec3 v3Pos2 = Geometry::CubicInterpolate(
							v3Start[i], 
							v3End[i], 
							v3VelA * fCurve[i], 
							v3VelB * fCurve[i], 
							((float)nSeg + 0.01f) / (float)k_nNumCornerSegs);	

						v3LightNormal = (v3Pos2 - v3Pos).CrossWithUnitY();
						v3LightNormal.Normalise();
					}
					else
					{
						v3Pos = 
							v3End[i] + 
							(v3End2[i] - v3End[i]) * 
							(float)(nSeg - k_nNumCornerSegs) / (float)k_nNumCornerSegs;
						v3LightNormal = (v3End2[i] - v3End[i]).CrossWithUnitY();
						v3LightNormal.Normalise();
					}
					static float fLastHeight = 0.0f;

					const float k_fFenceHeight = 2.0f;
					Collision collision = Physics::GetInstance().TestLineForCollision(
						v3Pos,
						-k_v3UnitY,
						1000.0f,
						Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);
					if (collision.CollisionOccurred())
					{
						v3Pos = collision.GetPosition() + k_v3UnitY * k_fFenceHeight;
						fLastHeight = v3Pos.y;
					}
					else
					{
						v3Pos.y = fLastHeight;
					}

					static float s_fLastDistance[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
					static Vec3 s_v3LastPos[4] = 
					{
						v3Start[0],
						v3Start[1],
						v3Start[2],
						v3Start[3]
					};
					s_fLastDistance[i] -= (v3Pos - s_v3LastPos[i]).GetMagnitude() * 0.1f;
					s_v3LastPos[i] = v3Pos;
					pCollisionObjectAABBMesh->AddVertex(v3Pos);
					float fLight = 255.0f * k_v3LightDir.y * 0.5f + 128.0f;
					int nLight = FastFloatToInt(fLight);
					Clamp(nLight, 0, 255);
					nLight = nLight + (nLight << 8) + (nLight << 16) + 0xFF000000;
					m_pMesh->AddVertex(v3Pos, Vec2(s_fLastDistance[i], 0.0f), nLight);
					fLight = 255.0f * v3LightNormal.Dot(k_v3LightDir) * 0.5f + 128.0f;
					nLight = FastFloatToInt(fLight);
					Clamp(nLight, 0, 255);
					nLight = nLight + (nLight << 8) + (nLight << 16) + 0xFF000000;
					m_pMesh->AddVertex(v3Pos, Vec2(s_fLastDistance[i], 0.0f), nLight);
					pCollisionObjectAABBMesh->AddVertex(v3Pos + v3BottomOffset);
					m_pMesh->AddVertex(v3Pos + v3BottomOffset, Vec2(s_fLastDistance[i], 3.0f), nLight);	
				}
			}
		}

				
		for (nStraight = 0; nStraight < nNumSegs; nStraight++)
		{			
			Straight& straight = m_pStraightArray[nStraight];
			
			Straight& straightA = m_pStraightArray[nStraight];
			Straight& straightB = nStraight + 1 < nNumSegs ? 
				m_pStraightArray[nStraight + 1] : m_pStraightArray[0];

			for (int nSeg = 0; nSeg < k_nNumCornerSegs * 2; nSeg++)
			{

				int nIndexOffset = (nStraight * (k_nNumCornerSegs * 2) + nSeg) * 8;
				int pnPolygon[4];
				int i;
				for (i = 0; i < 2; i++)
				{
					pnPolygon[0] = nIndexOffset + i * 4;
					pnPolygon[1] = nIndexOffset + 1 + i * 4;
					pnPolygon[2] = nIndexOffset + 9 + i * 4;
					pnPolygon[3] = nIndexOffset + 8 + i * 4;
					if (pnPolygon[0] >= nNumCollisionVertices)
						pnPolygon[0] -= nNumCollisionVertices;
					if (pnPolygon[1] >= nNumCollisionVertices)
						pnPolygon[1] -= nNumCollisionVertices;
					if (pnPolygon[2] >= nNumCollisionVertices)
						pnPolygon[2] -= nNumCollisionVertices;
					if (pnPolygon[3] >= nNumCollisionVertices)
						pnPolygon[3] -= nNumCollisionVertices;
					pCollisionObjectAABBMesh->AddPolygon(4, pnPolygon, s_nSurfaceAttribute);

					pnPolygon[0] = nIndexOffset + 10 + i * 4;
					pnPolygon[1] = nIndexOffset + 11 + i * 4;
					pnPolygon[2] = nIndexOffset + 3 + i * 4;
					pnPolygon[3] = nIndexOffset + 2 + i * 4;
					if (pnPolygon[0] >= nNumCollisionVertices)
						pnPolygon[0] -= nNumCollisionVertices;
					if (pnPolygon[1] >= nNumCollisionVertices)
						pnPolygon[1] -= nNumCollisionVertices;
					if (pnPolygon[2] >= nNumCollisionVertices)
						pnPolygon[2] -= nNumCollisionVertices;
					if (pnPolygon[3] >= nNumCollisionVertices)
						pnPolygon[3] -= nNumCollisionVertices;
					pCollisionObjectAABBMesh->AddPolygon(4, pnPolygon, s_nSurfaceAttribute);

					pnPolygon[0] = nIndexOffset + i * 4;
					pnPolygon[1] = nIndexOffset + 8 + i * 4;
					pnPolygon[2] = nIndexOffset + 10 + i * 4;
					pnPolygon[3] = nIndexOffset + 2 + i * 4;
					if (pnPolygon[0] >= nNumCollisionVertices)
						pnPolygon[0] -= nNumCollisionVertices;
					if (pnPolygon[1] >= nNumCollisionVertices)
						pnPolygon[1] -= nNumCollisionVertices;
					if (pnPolygon[2] >= nNumCollisionVertices)
						pnPolygon[2] -= nNumCollisionVertices;
					if (pnPolygon[3] >= nNumCollisionVertices)
						pnPolygon[3] -= nNumCollisionVertices;
					pCollisionObjectAABBMesh->AddPolygon(4, pnPolygon, s_nSurfaceAttribute);
				}	
				
				nIndexOffset = (nStraight * (k_nNumCornerSegs * 2) + nSeg) * 12;
				for (i = 0; i < 2; i++)
				{
					pnPolygon[0] = nIndexOffset + 1 + i * 6;
					pnPolygon[1] = nIndexOffset + 2 + i * 6;
					pnPolygon[2] = nIndexOffset + 14 + i * 6;
					pnPolygon[3] = nIndexOffset + 13 + i * 6;
					if (pnPolygon[0] >= nNumRenderVertices)
						pnPolygon[0] -= nNumRenderVertices;
					if (pnPolygon[1] >= nNumRenderVertices)
						pnPolygon[1] -= nNumRenderVertices;
					if (pnPolygon[2] >= nNumRenderVertices)
						pnPolygon[2] -= nNumRenderVertices;
					if (pnPolygon[3] >= nNumRenderVertices)
						pnPolygon[3] -= nNumRenderVertices;
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[1]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[3], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[1], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[3]);

					pnPolygon[0] = nIndexOffset + 16 + i * 6;
					pnPolygon[1] = nIndexOffset + 17 + i * 6;
					pnPolygon[2] = nIndexOffset + 5 + i * 6;
					pnPolygon[3] = nIndexOffset + 4 + i * 6;
					if (pnPolygon[0] >= nNumRenderVertices)
						pnPolygon[0] -= nNumRenderVertices;
					if (pnPolygon[1] >= nNumRenderVertices)
						pnPolygon[1] -= nNumRenderVertices;
					if (pnPolygon[2] >= nNumRenderVertices)
						pnPolygon[2] -= nNumRenderVertices;
					if (pnPolygon[3] >= nNumRenderVertices)
						pnPolygon[3] -= nNumRenderVertices;
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[1]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[3], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[1], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[3]);

					pnPolygon[0] = nIndexOffset + i * 6;
					pnPolygon[1] = nIndexOffset + 12 + i * 6;
					pnPolygon[2] = nIndexOffset + 15 + i * 6;
					pnPolygon[3] = nIndexOffset + 3 + i * 6;
					if (pnPolygon[0] >= nNumRenderVertices)
						pnPolygon[0] -= nNumRenderVertices;
					if (pnPolygon[1] >= nNumRenderVertices)
						pnPolygon[1] -= nNumRenderVertices;
					if (pnPolygon[2] >= nNumRenderVertices)
						pnPolygon[2] -= nNumRenderVertices;
					if (pnPolygon[3] >= nNumRenderVertices)
						pnPolygon[3] -= nNumRenderVertices;
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[1]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[3], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[1], pnPolygon[2]);
					m_pMesh->AddTriangle(pnPolygon[0], pnPolygon[2], pnPolygon[3]);
				}							
			}
		}

		m_pMesh->FinishedAddingGeometry();
		pCollisionObjectAABBMesh->FinishedAddingGeometry();

		m_pStaticObject = StaticObject::CreateNew();
		m_pStaticObject->Initialise(pCollisionObjectAABBMesh);
		pCollisionObjectAABBMesh->Release();
		Physics::GetInstance().AddStaticObject(m_pStaticObject);
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RaceTrack::Finalise()
{
	if (m_pStraightArray)
	{
		delete m_pStraightArray;
		m_pStraightArray = 0;
	}
		
	if (m_pMesh)
	{
		m_pMesh->Release();
		m_pMesh = 0;
	}
	if (m_pStaticObject)
	{
		Physics::GetInstance().RemoveStaticObject(m_pStaticObject);
		m_pStaticObject->Release();
		m_pStaticObject = 0;
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RaceTrack::Render()
{	
	if (m_pMesh)
		m_pMesh->Render();
}

//---------------------------------------------------------------------------------
// Assumes that the track has been added to the collision
//---------------------------------------------------------------------------------
MFrame RaceTrack::GetStartFrame(int nGridIndex)
{
	int nOffsetX = (nGridIndex & 3) - 1;
	int nOffset = nGridIndex;

	Straight& straight = m_pStraightArray[0];

	Vec3 v3Dir = straight.v3End - straight.v3Start;
	v3Dir.y = 0.0f;
	TA_ASSERT(v3Dir.GetMagnitude() > 0.0f);
	v3Dir.Normalise();

	Vec3 v3Start = straight.v3End;
	v3Start += (float)nOffset * -7.0f * v3Dir;
	Vec3 v3CrossDir = v3Dir.CrossWithUnitY();
	v3Start += (float)nOffsetX * -9.0f * v3CrossDir;

	Vec3 v3End = v3Start;
	v3End.y -= 250.0f;

	Collision collision = Physics::GetInstance().TestLineForCollision(
		v3Start,
		v3End,
		Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);
	if (collision.CollisionOccurred())
	{
		v3Start = collision.GetPosition();
		v3Start.y += 1.0f;
	}
	MFrame frame;
	frame.v3Translation = v3Start;
	frame.m33Rotation.SetToLookAt(
		v3Dir, 
		k_v3UnitY);

	return frame;
}
