//---------------------------------------------------------------------------------
// File Name: RacingAI.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "RacingAI.h"

#ifndef TA_DYNAMICOBJECTCAR_H
#include "../TA/Physics/DynamicObjectCar.h"
#endif // TA_DYNAMICOBJECTCAR_H

#ifndef TA_PHYSICS_H
#include "../TA/Physics/Physics.h"
#endif // TA_PHYSICS_H

#ifndef TA_RACETRACK_H
#include "RaceTrack.h"
#endif // TA_RACETRACK_H

using namespace TA;

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
RacingAI::RacingAI()
{
	m_pDynamicObjectCar = 0;
	m_nCurrentTrackSeg = 0;
	m_fSpeed = 0.0f;
	m_fTimeStuck = 0.0f;
}
//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------

RacingAI::~RacingAI()
{
	Finalise();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RacingAI::Initialise(DynamicObjectCar& physicsObjectCar, float fSpeed)
{
	if (m_pDynamicObjectCar)
	{
		TA_ASSERT(m_pDynamicObjectCar);
		Finalise();
	}
	m_pDynamicObjectCar = &physicsObjectCar;
	m_pDynamicObjectCar->AddRef();	
	m_fSpeed = fSpeed;
	m_fTimeStuck = 0.0f;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RacingAI::Finalise()
{
	if (m_pDynamicObjectCar)
	{
		m_pDynamicObjectCar->Release();
		m_pDynamicObjectCar = 0;
	}
	m_nCurrentTrackSeg = 0;
	m_fSpeed = 0.0f;
	m_fTimeStuck = 0.0f;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RacingAI::Update(float fDt, const RaceTrack* pRaceTrack)
{
	// Remove the car if it is stuck
	if (!m_pDynamicObjectCar)
		return;
	if (m_pDynamicObjectCar->GetLinearVelocity().GetMagnitude() < 1.0f)
		m_fTimeStuck += fDt;
	else
		m_fTimeStuck = 0.0f;
	if (m_fTimeStuck > 2.0f)
	{
		Physics::GetInstance().RemoveDynamicObject(m_pDynamicObjectCar);
		m_pDynamicObjectCar->Release();
		m_pDynamicObjectCar = 0;
		return;
	}

	TA_ASSERT(m_pDynamicObjectCar);
	AdvanceSegment(pRaceTrack);

	const MFrame& carFrame = m_pDynamicObjectCar->GetFrame();

	const RaceTrack::Straight& straight = pRaceTrack->GetStraight(m_nCurrentTrackSeg);
	const RaceTrack::Straight& nextStraight = 
		m_nCurrentTrackSeg + 1 < pRaceTrack->GetNumSegments() ?
		pRaceTrack->GetStraight(m_nCurrentTrackSeg + 1) :
		pRaceTrack->GetStraight(0);

	Vec3 v3Normal = straight.v3End - straight.v3Start;
	TA_ASSERT(v3Normal.GetMagnitude() != 0.0f);
	v3Normal.Normalise();
	Vec3 v3CrossNormal = -v3Normal.CrossWithUnitY(); 

	Vec3 v3NextNormal = nextStraight.v3End - nextStraight.v3Start;
	TA_ASSERT(v3NextNormal.GetMagnitude() != 0.0f);
	v3NextNormal.Normalise();
	Vec3 v3NextCrossNormal = -v3NextNormal.CrossWithUnitY(); 
	
	float fCenterDistance = v3CrossNormal.Dot(straight.v3End - carFrame.v3Translation);
	float fEndDistance = v3Normal.Dot(straight.v3End - carFrame.v3Translation);

	float fSinCornerAngle = v3Normal.Cross(v3NextNormal).y; // The cross product is a waste if we are only using y

	Vec3 v3TargetPos;
	float fOverSteer = 0.0f;

	const float k_fEarlyCorneringDistance = 30.0f;
	const float k_fSpeedNormalisation = 3.6f / 220.0f;

	float fNormalSpeed = 
		k_fSpeedNormalisation *
		m_pDynamicObjectCar->GetLinearVelocity().Dot(v3Normal);
	if (fNormalSpeed < 0.0f)
		fNormalSpeed = 0.0f;


	fSinCornerAngle *= fNormalSpeed;
	
	float fBrake = 0.0f;

	if (fEndDistance - k_fEarlyCorneringDistance * Fabs(fSinCornerAngle) - 
		fCenterDistance * fSinCornerAngle * 1.0f < 0.0f)
	{
		v3TargetPos = 
			nextStraight.v3End + 
			v3NextNormal * 100.0f + 
			v3NextCrossNormal * -25.0f * Sign(fSinCornerAngle) * (m_fSpeed - 0.5f);
		fBrake = 0.2f;

	}
	else
	{
		v3TargetPos = 
			straight.v3End + 
			v3Normal * 100.0f + 
			v3CrossNormal * -15.0f * Sign(fSinCornerAngle) * (m_fSpeed - 0.5f);

		fOverSteer = -m_pDynamicObjectCar->GetLinearVelocity().Dot(v3CrossNormal) * 0.04f;
		fOverSteer = Fabs(fOverSteer) * fOverSteer;
		Clamp(fOverSteer, -0.5f, 0.5f);
		 
	}

	Vec3 v3Direction = (v3TargetPos - carFrame.v3Translation);
	v3Direction.y = 0.0f;
	TA_ASSERT(v3Direction.GetMagnitude() != 0.0f); // Assume that this case has been avoided earlier
	v3Direction.Normalise();


	v3Direction *= carFrame.m33Rotation.GetTransposeAsInverse();

	float fSteeringAngle = ATan2(v3Direction.x, v3Direction.z);
	fSteeringAngle = fSteeringAngle * fSteeringAngle * fSteeringAngle * 40.0f;

	
	fSteeringAngle -= m_pDynamicObjectCar->GetAngularVelocity().y * 0.4f;
	fSteeringAngle += fOverSteer;
	Clamp(fSteeringAngle, -1.0f, 1.0f);


	m_pDynamicObjectCar->UpdateControls(
		fSteeringAngle,
		m_fSpeed,			// Accelerator
		0.0f,
		fBrake,
		1.0f,
		1.0f);
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RacingAI::AdvanceSegment(const RaceTrack* pRaceTrack)
{
	const float k_fGotoNextSegmentDistance = 1.0f;
	TA_ASSERT(m_pDynamicObjectCar);
	for (;;)
	{
		const RaceTrack::Straight& straight = pRaceTrack->GetStraight(m_nCurrentTrackSeg);
		Vec3 v3Normal = straight.v3End - straight.v3Start;
		TA_ASSERT(v3Normal.GetMagnitude() != 0.0f);
		v3Normal.Normalise();
		if (v3Normal.Dot(m_pDynamicObjectCar->GetPosition() - straight.v3End) < -k_fGotoNextSegmentDistance)
			break;

		m_nCurrentTrackSeg++;
		if (m_nCurrentTrackSeg >= pRaceTrack->GetNumSegments())
			m_nCurrentTrackSeg = 0;
	}
}
