//---------------------------------------------------------------------------------
// File Name: DemoTerrain.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "../Interface/IApp.h"
#include "../Interface/IDemoPlugin.h"
#include "../Interface/IMouse.h"
#include "../Interface/ICamera.h"

#include "../TA/Common/RandFunc.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/CollisionObjectCombo.h"

#include "../Shared/Landscape.h"
#include "../Shared/DemoPhysicsRender.h"

#include <windows.h>

using namespace TA;
	
static bool KeyIsDown(int nKey) { return (GetAsyncKeyState(nKey) & 0x8000) != 0; }
static bool KeyIsJustPressed(int nKey) { return (GetAsyncKeyState(nKey) & 0x8001) == 0x8001; }

class DemoTerrain : public IDemoPlugin
{
public:
	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DemoTerrain()
	{
		m_nNumDynamicObjects = 0;
		m_ppDynamicObjectList = 0;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	~DemoTerrain()
	{
		Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Release() { delete this; }

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Initialise(IApp* pApp)
	{
		m_pApp = pApp;

		// Implement physics debug rendering functions.
		DemoPhysicsRender::Initialise(pApp);

		// Initialise mouse.
		m_pMouse = m_pApp->CreateMouse();
		m_pMouse->Initialise(true);

		// Initialise camera.
		m_pCamera = m_pApp->CreateCamera();
		m_pCamera->Initialise();

		// Initialise Physics.
		AABB worldAABB;
		worldAABB.Initialise(k_v3Zero, Vec3(5000.0f, 5000.0f, 5000.0f));
		Physics::CreateInstance(worldAABB);	
		Physics::GetInstance().SetGravity(-1.0f * 9.81f * k_v3UnitY);
	//	Physics::GetInstance().SetRenderCollisionsEnabled(true);



		//---------------------------------------------------------------------------------
		// Initialise Landscape
		//---------------------------------------------------------------------------------

		// Initialise Landscape collision object.
		m_pLandscape = new Landscape();
		m_pLandscape->Initialise(m_pApp, 6.0f);

		// Create a static object for the collision.
		StaticObject* pStaticObject = StaticObject::CreateNew();
		pStaticObject->Initialise(m_pLandscape->GetCollisionObject());

		// Add the static object to Physics.
		Physics::GetInstance().AddStaticObject(pStaticObject);
		// If we wanted to keep call Physics::GetInstance().RemoveStaticObject(pStaticObject) 
		// later, we would have to keep a pointer to pStaticObject.
		pStaticObject->Release(); 
		pStaticObject = 0;



		//---------------------------------------------------------------------------------
		// Initialise Dynamic Objects
		//---------------------------------------------------------------------------------		
		
		// Initialise objects
		m_nNumDynamicObjects = 32;
		m_ppDynamicObjectList = new DynamicObject* [m_nNumDynamicObjects];
		for (int nDynamicObject = 0; nDynamicObject < m_nNumDynamicObjects; nDynamicObject++)
		{			
			
			// Initialise collision object
			CollisionObjectCombo* pCollisionObjectCombo = CollisionObjectCombo::CreateNew();
	
			const int k_nNumPoints = 50;
			Vec3 pv3PointList[k_nNumPoints];
			const float k_fRandScale = 0.75f;
			for (int i = 0; i < k_nNumPoints; i++)
			{
				Vec3 v3Test(
					RandFunc::RandBetweenPosNeg(1.0f),
					RandFunc::RandBetweenPosNeg(1.0f),
					RandFunc::RandBetweenPosNeg(1.0f));
				float fMag = v3Test.GetMagnitude();
				if (fMag > 0.0f)
					v3Test /= fMag;
				else
					v3Test = k_v3UnitY;
				pv3PointList[i] = v3Test * RandFunc::RandBetween(0.5f, 1.0f) * k_fRandScale;
			}

			pCollisionObjectCombo->InitialiseFromPointList(
				pv3PointList, 
				k_nNumPoints);

			const float k_fRandRange = 50.0f;
			Vec3 v3GroundPos(
				RandFunc::RandBetweenPosNeg(k_fRandRange), 
				10.0f, 
				RandFunc::RandBetweenPosNeg(k_fRandRange));

			// Get placement height
			Collision collision = Physics::GetInstance().TestLineForCollision(
				v3GroundPos + k_v3UnitY * 1000.0f,
				-k_v3UnitY * 2000.0f,
				Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

			if (collision.CollisionOccurred())
				v3GroundPos = collision.GetPosition();
			v3GroundPos.y += 10.0f;

			m_ppDynamicObjectList[nDynamicObject] = DynamicObject::CreateNew();
			TA_ASSERT(m_ppDynamicObjectList[nDynamicObject]);
			DynamicObject& dynamicObject = *m_ppDynamicObjectList[nDynamicObject];
			dynamicObject.Initialise(pCollisionObjectCombo);
			dynamicObject.SetFrame(
				MFrame(
					v3GroundPos,
					Mat33(Mat33::ROTATION_AXIS_Y, k_fPi * 0.0f * nDynamicObject)));
			v3GroundPos.y += 0.98f;
			dynamicObject.SetFriction(3.0f);
			Physics::GetInstance().AddDynamicObject(&dynamicObject);	

			pCollisionObjectCombo->Release();

		}
		
		m_bHelp = true;
		m_bInvertMouse = true;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Finalise()
	{
		if (m_ppDynamicObjectList)
		{
			for (int i = 0; i < m_nNumDynamicObjects; i++)
			{
				if (m_ppDynamicObjectList[i])
				{
					m_ppDynamicObjectList[i]->Release();
					m_ppDynamicObjectList[i] = 0;
				}
			}
			delete []m_ppDynamicObjectList;
			m_ppDynamicObjectList = 0;
		}

		if (m_pLandscape)
		{
			delete m_pLandscape;
			m_pLandscape = 0;
		}
		if (m_pMouse)
		{
			m_pMouse->Release();
			m_pMouse = 0;
		}
		if (m_pCamera)
		{
			m_pCamera->Release();
			m_pCamera = 0;
		}
		m_pApp = 0;

		Physics::DestroyInstance();

		DemoPhysicsRender::Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Update(float fDt)
	{
		// Update Input
		m_pMouse->Update();
		if (m_pMouse->IsLeftButtonDown())
		{		
			TA_ASSERT(m_ppDynamicObjectList[0]);
			Physics::GetInstance().SetDynamicObjectIsMoving(m_ppDynamicObjectList[0]);
			m_ppDynamicObjectList[0]->ApplyImpulse(
				m_pCamera->GetRotation().v3Z * 7.5f,			// Impulse.
				m_ppDynamicObjectList[0]->GetPosition());	// World position.
		}

		if (KeyIsJustPressed(VK_F1))
			m_bHelp = !m_bHelp;
		if (KeyIsJustPressed('I'))
			m_bInvertMouse = !m_bInvertMouse;


		// Update physics
		Physics::GetInstance().Update(fDt);

		// Update the camera.	
		m_pCamera->SetNearClip(0.1f);
		m_pCamera->SetFarClip(1000.0f);
		m_pCamera->SetViewDistance(10.0f);
		m_pCamera->UpdateRotation(
			m_pMouse->GetXMovement(),
			m_bInvertMouse ? m_pMouse->GetYMovement() : -m_pMouse->GetYMovement());
		TA_ASSERT(m_ppDynamicObjectList[0]);
		m_pCamera->SetTargetPosition(
			m_ppDynamicObjectList[0]->GetPosition() + k_v3UnitY);
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Render()
	{
		// Set the camera
		m_pCamera->Realise();
		
		// Render physics (renders collision objects)
		Physics::GetInstance().Render();

		// Render the Landscape
		m_pLandscape->Render();

		RenderHelp();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void RenderHelp()
	{
		if (m_bHelp)
		{
			m_pApp->RenderText("Toggle help on and off (F1)");
			m_pApp->RenderText("Move the mouse to rotate the camera");
			m_pApp->RenderText("Invert mouse (I)");
			m_pApp->RenderText("Move object (Left Mouse Button)");
			m_pApp->RenderText("Quit (Esc)");
		}
	}

private:	
	IApp* m_pApp;
	IMouse* m_pMouse;	
	ICamera* m_pCamera;
	Landscape* m_pLandscape;

	int m_nNumDynamicObjects;
	DynamicObject** m_ppDynamicObjectList;
	bool m_bHelp;
	bool m_bInvertMouse;
};

TA_IMPLEMENT_DEMO_PLUGIN(DemoTerrain);