//---------------------------------------------------------------------------------
// File Name: Landscape.cpp
// Description:
//
// Copyright (C) 2002 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "Landscape.h"

#include "../Interface/IApp.h"
#include "../Interface/IMesh.h"

//#include "../TA/Common/Error.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"

#include "../Shared/HeightField.h"

using namespace TA;

struct Landscape::Data
{
	HeightField* pHeightField;
	IMesh* pMesh;
	CollisionObjectAABBMesh* pCollisionObject;
};

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
Landscape::Landscape()
{
	m_pData = 0;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
Landscape::~Landscape()
{
	Finalise();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void Landscape::Initialise(IApp* pApp, float fVScale)
{
	if (m_pData)
	{
		TA_ASSERT(0);
		Finalise();
	}
	m_pData = new Data;
	m_pData->pMesh = 0;
	m_pData->pCollisionObject = 0;

	const float k_fLandscapeHorizontalScale = 16.0f; 
	float k_fLandscapeVerticalScale = 150.0f * fVScale;
	const float k_fLandscapeTextureScale = 0.8f;

	// Create fractal landscape
	const int k_nNumLandscapeDetailLevels = 7;
	m_pData->pHeightField = new HeightField();
	m_pData->pHeightField->Initialise(
		k_nNumLandscapeDetailLevels, 
		k_nNumLandscapeDetailLevels, 
		1.95f,		// fractal dimension
		2,			// Random seed
		k_fLandscapeHorizontalScale,
		k_fLandscapeVerticalScale);	

	m_pData->pMesh = pApp->CreateMesh();	
	m_pData->pCollisionObject = CollisionObjectAABBMesh::CreateNew();
	
	const int nNumVertices = 
		m_pData->pHeightField->GetSizeX() * 
		m_pData->pHeightField->GetSizeY();	
	const int nNumIndices = 
		(m_pData->pHeightField->GetSizeX() - 1) * 
		(m_pData->pHeightField->GetSizeY() - 1) * 6;	

	m_pData->pMesh->Initialise(nNumVertices, nNumIndices, "Assets/Textures/Landscape.tga");
	m_pData->pCollisionObject->Initialise(nNumVertices, nNumIndices / 3, nNumIndices);

	// Set vertices
	int nY;
	for (nY = 0; nY < m_pData->pHeightField->GetSizeY(); nY++)
	{
		for (int nX = 0; nX < m_pData->pHeightField->GetSizeX(); nX++)
		{
			Vec3 v3Pos(
				(float)(nX - (m_pData->pHeightField->GetSizeX() >> 1)) * k_fLandscapeHorizontalScale,
				m_pData->pHeightField->GetHeightAt(nY, nX) * k_fLandscapeVerticalScale - 3.0f,
				(float)(nY - (m_pData->pHeightField->GetSizeY() >> 1)) * k_fLandscapeHorizontalScale);
			Vec2 v2T0(
				nX * k_fLandscapeTextureScale,
				nY * k_fLandscapeTextureScale);

			// Give a bit of colour variation.
			const float k_fScale = 8.0f;
			float fIntensity = 
				(Sin(nX * 0.31f * k_fScale) + 1.0f) *
				(Sin((nY * 0.31f + 0.4f) * k_fScale) + 1.0f) * 0.2f +
				(Sin(nX * 0.11f * k_fScale) + 1.0f) *
				(Sin(nY * 0.11f * k_fScale) + 2.0f) * 0.4f +
				(Sin((nX * 0.17f + 0.3f) * k_fScale) + 1.0f)*
				(Sin(nY * 0.17f * k_fScale) + 1.0f) * 0.3f  +
				(Sin(nX * 0.013f * k_fScale) + 2.0f) *
				(Sin(nY * 0.013f * k_fScale) + 1.0f) * 0.5f +
				(Sin(nX * 0.07f * k_fScale) + 2.0f) *
				(Sin(nY * 0.07f * k_fScale) + 1.0f) * 0.9f +
				(Sin(nX * 0.03f * k_fScale) + 2.0f) *
				(Sin(nY * 0.03f * k_fScale) + 1.0f) * 1.4f;
			fIntensity /= 32.0f;
			fIntensity += 0.5f;
			int nIntensity = (int)(fIntensity * 255.0f);
			Clamp(nIntensity, 0, 255);
			m_pData->pMesh->AddVertex(v3Pos, v2T0, 0xff000000 + (nIntensity << 16) + (nIntensity << 8) + nIntensity);
			m_pData->pCollisionObject->AddVertex(v3Pos);	
		}
	}
	
    // Set Indices
	for (nY = 0; nY < m_pData->pHeightField->GetSizeY() - 1; nY++)
	{
		for (int nX = 0; nX < m_pData->pHeightField->GetSizeX() - 1; nX++)
		{
			int pnIndicesA[3] = 
			{
				(nY) * m_pData->pHeightField->GetSizeX() + nX,
				(nY) * m_pData->pHeightField->GetSizeX() + nX + 1,
				(nY + 1) * m_pData->pHeightField->GetSizeX() + nX + 1,
			};
			m_pData->pMesh->AddTriangle(pnIndicesA[0], pnIndicesA[2], pnIndicesA[1]);
			m_pData->pCollisionObject->AddPolygon(3, pnIndicesA);

			int pnIndicesB[3] = 
			{
				(nY) * m_pData->pHeightField->GetSizeX() + nX,
				(nY + 1) * m_pData->pHeightField->GetSizeX() + nX + 1,
				(nY + 1) * m_pData->pHeightField->GetSizeX() + nX,
			};
			m_pData->pMesh->AddTriangle(pnIndicesB[0], pnIndicesB[2], pnIndicesB[1]);
			m_pData->pCollisionObject->AddPolygon(3, pnIndicesB);
		}
	}
	m_pData->pMesh->FinishedAddingGeometry();
	m_pData->pCollisionObject->FinishedAddingGeometry();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void Landscape::Finalise()
{
	if (m_pData)
	{
		if (m_pData->pHeightField)
		{
			delete m_pData->pHeightField;
			m_pData->pHeightField = 0;
		}
		if (m_pData->pMesh)
		{
			m_pData->pMesh->Release();
			m_pData->pMesh = 0;
		}
		if (m_pData->pCollisionObject)
		{
			m_pData->pCollisionObject->Release();
			m_pData->pCollisionObject = 0;
		}
	
		delete m_pData;
		m_pData = 0;
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void Landscape::Render()
{
	TA_ASSERT(m_pData);
	TA_ASSERT(m_pData->pMesh);
	m_pData->pMesh->Render();

	
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
CollisionObjectAABBMesh* Landscape::GetCollisionObject()
{
	TA_ASSERT(m_pData);
	TA_ASSERT(m_pData->pCollisionObject);
	return m_pData->pCollisionObject;
}
